# A module to nicely deal with Debian packages.  Well, fairly nicely :)
#
# Certainly some of this stuff could be better done..

import os
import sys
import string
import re

BadArchiveException="BadArchiveException"
BadControlFieldException="BadControlFieldException"
PackageNotFoundException="PackageNotFoundException"
NotRootException="NotRootException"
PatchException="PatchException"
BadVersionException="BadVersionException"
BadArgException="BadArgException"

ctrfieldre=re.compile('([^:]*):(.*)')
dpkgrepackre=re.compile('^dpkg-deb: building package `[^\']*\' in `./([^\']*)\'')
dpkgdebre=re.compile('^dpkg-deb: building package `[^\']*\' in `([^\']*)\'')

# This is used as a magic-number to indicate the format version of the diffiles
diffver='2.0.diff-0.2'

def parsecontrol(file):
    control={}

    line=file.readline()
    linenum=1
    while line <> "":
	found=ctrfieldre.match(line)
	if not found:
	    raise BadControlFieldException,'Syntax error in control file at line %i: %s'%(linenum,line)

	field=string.strip(found.group(1))
	value=string.strip(found.group(2))

	if field == "Description":
	    control[field]=value
	    value=''
	    field="Description-Long"

	line=file.readline()

	linenum=linenum+1

	while line <> '' and line[0] == ' ':
	    value=value+line
	    line=file.readline()
	    linenum=linenum+1

	control[field]=value

    return control

def parsemoves(file):
    moves=[]
    line=file.readline()
    while line <> '':
	linelist=string.split(line)
	if len(linelist) <= 3 or len(linelist)%2 == 0:
	    raise BadArchiveException,'Wrong number of items in movesfile: %s'%line

	group=linelist[:3]
	group[1]=int(group[1])
	# Change the size to an integer

	linelist=linelist[3:]
	dests=[]

	while linelist <> []:
	    dests.append(linelist[0],int(linelist[1]))
	    linelist=linelist[2:]

	moves.append(tuple(group)+(dests,))

	line=file.readline()
    if file.close():
	raise BadArchiveException,'Couldn\'t read movesfile'
    return moves

hex='0123456789abcdef'

def md5totext(md5sum):
    retval=""
    for i in md5sum:
	retval=retval+hex[ord(i)%16]+hex[ord(i)/16]
    return retval

class debfile:
    def __init__(self,filename):
	# FIXME: do this better :)
	if os.system("ar t %s | grep ^debian-binary > /dev/null 2>&1"%filename):
	    raise BadArchiveException,"%s does not look like a Debian archive"%filename

	self.filename=os.path.abspath(filename)

    def __getattr__(self,attr):
	if attr == 'contents':
	    dpkgrun=os.popen('dpkg-deb --fsys-tarfile %s | tar tf -'%self.filename)
	    self.contents=string.split(dpkgrun.read())
	    if dpkgrun.close():
		if hasattr(self.__dict__,'contents'):
		    del self.contents
		raise BadArchiveException,'dpkg failed to extract the contents of %s'%filename
	    return self.contents
	elif attr == 'conffiles':
	    if os.system('ar p %s control.tar.gz | tar ztf - | grep -q conffiles'%self.filename):
		self.conffiles=[]
	    else:
		conffiles=os.popen('ar p %s control.tar.gz | tar zxOf - conffiles'%self.filename)
		self.conffiles=string.split(conffiles.read())

		if conffiles.close():
		    if haskey(self.__dict__,'conffiles'):
			del self.conffiles
		    raise BadArchiveException,'Failed to read the conffiles list for %s'%self.filename

	    return self.conffiles
	elif attr == 'control':
	    # Get information about the package:
	    dpkgrun=os.popen('dpkg-deb -f %s'%self.filename)
	    self.control=parsecontrol(dpkgrun)
	    if dpkgrun.close():
		if hasattr(self.__dict__,'control'):
		    del self.control
		raise BadControlFieldException,'dpkg failed to read the control file of %s'%filename
	    return self.control
	elif attr == 'size':
	    self.size=os.path.getsize(self.filename)

	    return self.size
	else:
	    raise AttributeError,attr

    def __repr__(self):
	return 'Debian package %s version %s'%(self['Package'],self['Version'])

    def extractdata(self,destdir="."):
	if os.system("dpkg-deb -x %s %s"%(self.filename,destdir)):
	    raise BadArchiveException,"Couldn't extract %s"%self.filename

    def extractcontrol(self,destdir="."):
	# Mostly not needed with the automagic control extraction the object
	# itself does.
	if os.system("dpkg-deb -e %s %s"%(self.filename,destdir)):
	    raise BadArchiveException,"Couldn't extract control information from %s"%self.filename

    # Mapping emulation:
    def has_key(self,key):
	return self.control.has_key(key)

    def __len__(self):
	return len(self.control)

    def __getitem__(self,key):
	return self.control[key]

def builddeb(dir,dest=None):
    if dest == None:
	dest=os.path.basename(dir)+'.deb'

    proc=os.popen('dpkg-deb -b %s %s'%(dir,dest))
    file=None
    line=proc.readline()
    while line <> '':
	found=dpkgdebre.match(line)
	if found:
	    file=found.group(1)
	line=proc.readline()
    if proc.close() or not found:
	raise BadArchiveException,"%s could not be turned into a Debian package"%dir

    return debfile(file)

def rebuilddeb(pkgname,where='.'):
    """Rebuilds a Debian package already on the system, places it in the
    given directory"""

    if os.getuid() <> 0:
	raise NotRootException,"You must be root (or fakeroot) to rebuild packages"

    lastdir=os.getcwd()
    os.chdir(where)
    process=os.popen('dpkg-repack %s'%pkgname)
    os.chdir(lastdir)

    line=process.readline()
    filename=None
    while line <> "":
	found=dpkgrepackre.match(line)
	if found:
	    filename='%s/%s'%(where,found.group(1))

	line=process.readline()

    retval=process.close()

    if retval:
	raise PackageNotFoundException,"Package %s not found on system"%pkgname

    if filename == None:
	raise PackageNotFoundException,"Couldn't parse dpkg-repack output while repacking %s"%pkgname

    return debfile(filename)

class debdiff:
    def __init__(self,filename):
	contents=os.popen('ar t %s'%filename)

	if contents.read() <> 'debian-diff\ncontrol.tar.gz\ndata.tar.gz\ndelta.tar.gz\nmoves\n':
	    contents.close()
	    raise BadArchiveException,"%s doesn't look like a Debian binary patch"%filename

	if contents.close():
	    raise BadArchiveException,"%s doesn't look like a Debian binary patch (ar complained)"%filename

	version=os.popen('ar p %s debian-diff'%filename).readline()
	if string.strip(version) <> diffver:
	    raise BadArchiveException,"%s is the wrong patch format version (expected %s, got %s)"%(filename,diffver,version)

	self.filename=os.path.abspath(filename)

    def extractdata(self,destdir='.'):
	if os.system('ar p %s data.tar.gz | tar -zxf - -C %s'%(self.filename,destdir)):
	    raise BadArchiveException,'Failed to extract data.tar.gz from %s'%self.filename

    def extractcontrol(self,destdir='.'):
	if os.system('ar p %s control.tar.gz | tar -zxf - -C %s'%(self.filename,destdir)):
	    raise BadArchiveException,'Failed to extract control.tar.gz from %s'%self.filename

    def extractdeltas(self,destdir='.'):
	if os.system('ar p %s delta.tar.gz | tar -zxf - -C %s'%(self.filename,destdir)):
	    raise BadArchiveException,'Failed to extract delta.tar.gz from %s'%self.filename

    def apply(self,file,debdest='.',diffdest=None):
	"""Applies this patch to a debfile object and returns a tuple
	(newfile,reverse) which contains the new debfile and a debdiff which
	can be used to recreate the original file from the new one"""

	if diffdest==None:
	    if os.path.isdir(debdest):
		diffdest=debdest
	    else:
		raise BadArgException,'If debdest is a filename, diffdest must also be supplied'

	if self['Patches'] <> file['Version']:
	    raise BadVersionException,'Version mismatch: seen %s, expected %s'%(file['Version'],self['Patches'])

	import tempfile
	import shutil
	dir=tempfile.mktemp()

	# Create a working directory
	os.mkdir(dir)

	try:
	    olddir='%s/old'%dir
	    newdir='%s/new'%dir
	    deltadir='%s/delta'%dir

	    os.mkdir(olddir)
	    os.mkdir(newdir)
	    os.mkdir(deltadir)

	    file.extractdata(olddir)

	    self.extractdata(newdir)
	    self.extractdeltas(deltadir)

	    for (fromfile,size,md5sum,dests) in self.moves:
		oldfile='%s/data/%s'%(olddir,fromfile)
		if size <> os.path.getsize(oldfile):
		    raise PatchException,'%s is the wrong size (expected %i, got %i)'%(fromfile,size,os.path.getsize(oldfile))

		if md5sum <> md5totext(md5.new(open(oldfile).read()).digest()):
		    raise PatchException,'md5sum mismatch for %s, has it been modified?'%fromfile

		for (tofile,tomode) in dests:
		    shutil.copy('%s/data/%s'%(olddir,fromfile),'%s/data/%s'%(newdir,tofile))
		    os.chmod(oldfile,tomode)

	    for i in self.deltas:
		# Not efficient to work out the deltas twice? ..maybe
		# extractdeltas should use tar's -v option and extract
		# a delta list from that?  Or I could just do this the right
		# way :)
		deltafile='%s/%s'%(deltadir,i)
		if os.path.isfile(deltafile):
		    # delta.tar.gz should only contain files and directories..
		    mode=os.stat(deltafile)[0]
		    if os.system('xdelta patch %s %s/%s %s/%s'%(deltafile,olddir,i,newdir,i)):
			raise PatchException,"Couldn't patch %s"%i
		    os.chmod('%s/%s'%(newdir,i),mode)
		    # Preserve [new] permissions

	    # Ok, now create the control area:
	    os.mkdir('%s/DEBIAN'%newdir,0755)
	    self.extractcontrol('%s/DEBIAN'%newdir)

	    newdeb=builddeb(newdir,debdest)

	    # FIXME: we can do this more efficiently by using what we know
	    #       while performing the other operations!  [ note that that
	    #       would be less reliable, though, and may not be worth the
	    #       trouble ]

	    newdiff=builddiff(newdeb,file,diffdest)
	finally:
	    shutil.rmtree(dir)

	return (newdeb,newdiff)

    def __getattr__(self,attr):
	if attr == 'control':
	    controlfile=os.popen('ar p %s control.tar.gz | tar zxOf - control'%self.filename)

	    self.control=parsecontrol(controlfile)

	    if controlfile.close():
		if haskey(self.__dict__,'control'):
		    del self.control
		raise BadControlFieldException,'Failed to read the control file of %s'%self.filename

	    return self.control
	elif attr == 'conffiles':
	    if os.system('ar p %s control.tar.gz | tar ztf - | grep -q conffiles'):
		self.conffiles=[]
	    else:
		conffiles=os.popen('ar p %s control.tar.gz | tar zxOf - conffiles'%self.filename)
		self.conffiles=string.split(conffiles.read())

		if conffiles.close():
		    if haskey(self.__dict__,'conffiles'):
			del self.conffiles
		    raise BadArchiveException,'Failed to read the conffiles list for %s'%self.filename

	    return self.conffiles
	elif attr == 'contents':
	    # Use the contents of data.tar.gz as the contents again
	    contents=os.popen('ar p %s data.tar.gz | tar ztf -'%self.filename)
	    self.contents=string.split(contents.read())

	    if contents.close():
		if haskey(self.__dict__,'contents'):
		    del self.contents
		raise BadArchiveException,'Failed to read the file list of %s'%self.filename

	    return self.contents
	elif attr == 'deltas':
	    deltas=os.popen('ar p %s delta.tar.gz | tar ztf -'%self.filename)
	    self.deltas=string.split(deltas.read())

	    if deltas.close():
		if haskey(self.__dict__,'deltas'):
		    del self.deltas
		raise BadArchiveException,'Failed to read the delta list of %s'%self.filename

	    return self.deltas
	elif attr == 'moves':
	    # Moves are returned in the format:
	    # (fromfile,size,md5sum,dests)
	    # where dests is a list of names.  Maybe I should start using a
	    # class for this?
	    moves=os.popen('ar p %s moves'%self.filename)
	    self.moves=parsemoves(moves)
	    return self.moves
	elif attr == 'size':
	    self.size=os.path.getsize(self.filename)

	    return self.size
	else:
	    raise AttributeError,attr

    def __repr__(self):
	return 'Debian binary diff from version %s to %s of %s'%(self['Patches'],self['Version'],self['Package'])

    # Mapping emulation:
    def has_key(self,key):
	return self.control.has_key(key)

    def __len__(self):
	return len(self.control)

    def __getitem__(self,key):
	return self.control[key]

def builddiff(fromfile,tofile,dest='.'):
    """Returns a debdiff object referring to a file which contains a diff
    between fromfile and tofile."""
    import tempfile
    import shutil
    import md5

    dir=tempfile.mktemp()
    os.mkdir(dir)

    try:
	fromdir='%s/from'%dir
	todir='%s/to'%dir

	for i in [fromdir,'%s/data'%fromdir,'%s/control'%fromdir,todir,'%s/data'%todir,'%s/shipping'%todir,'%s/control'%todir,'%s/delta'%todir]:
	    os.mkdir(i)

	for i in [(fromfile,fromdir),(tofile,todir)]:
	    i[0].extractdata('%s/data'%i[1])
	    i[0].extractcontrol('%s/control'%i[1])

	# Now I clone directories *and* snarf md5sums of files (to find when
	# files have moved)

	tofiles=[]
	fromfiles=[]

	for i in fromfile.contents:
	    filename='%s/data/%s'%(fromdir,i)
	    if os.path.isfile(filename):
		fromfiles.append( (md5.new(open(filename).read()).digest(),i,os.path.getsize(filename)))
	fromfiles.sort(lambda x,y:cmp(x[0],y[0]))

	for i in tofile.contents:
	    filename='%s/data/%s'%(todir,i)
	    if os.path.isfile(filename):
		tofiles.append( (md5.new(open(filename).read()).digest(),i,os.path.getsize(filename)))
	    elif i <> './' and os.path.isdir(filename):
		mode=os.stat(filename)[0]

		# Populate directories (may not always be needed but it makes
		# my life easier :) )
		os.mkdir('%s/shipping/%s'%(todir,i))
		os.mkdir('%s/delta/%s'%(todir,i))
	tofiles.sort(lambda x,y:cmp(x[0],y[0]))

	moves=[]
	while tofiles <> [] and fromfiles <> []:
	    (tomd5sum,tofilename,tosize)=tofiles.pop()
	    (frommd5sum,fromfilename,fromsize)=fromfiles.pop()
	    while frommd5sum > tomd5sum:
		(frommd5sum,fromfilename,fromsize)=fromfiles.pop()
	    while tomd5sum > frommd5sum:
		(tomd5sum,tofilename,tosize)=tofiles.pop()

	    moved=0
	    foundlist=[]
	    while frommd5sum == tomd5sum:
		if fromsize == tosize:
		    print 'Old %s is identical to new %s'%(fromfilename,tofilename)
		    foundlist.append((tofilename,os.stat('%s/data/%s'%(todir,tofilename))[0]))
		    os.remove('%s/data/%s'%(todir,tofilename))
		    # We don't want to accidentally make a delta for it or include
		    # it in the final patch..
		    moved=1

		(tomd5sum,tofilename,tosize)=tofiles.pop()

	    if moved:
		moves.append((fromfilename,fromsize,md5totext(frommd5sum),foundlist))
		os.remove('%s/data/%s'%(fromdir,fromfilename))

	for i in tofile.conffiles:
	    os.link('%s/data/%s'%(todir,i),'%s/shipping/%s'%(todir,i))
	    os.remove('%s/data/%s'%(todir,i))

	# Now make deltas for files which are in the same place and unmoved, and
	# move new files to shipping
	class walk:
	    def __init__(self,sub,fromdir,todir):
		# I cannot use the stock os.path.walk() function easily; I need to
		# know the *relative* path names.
		for file in os.listdir('%s/data/%s'%(todir,sub)):
		    topath='%s/data/%s/%s'%(todir,sub,file)
		    # This will always exist..
		    frompath='%s/data/%s/%s'%(fromdir,sub,file)
		    if os.path.isdir(topath):
			# Recurse.
			self.__init__('%s/%s'%(sub,file),fromdir,todir)
		    elif os.path.islink(topath) or not os.path.exists(frompath):
			# Move it straight to shipping (link to preserve mode)
			# I move symlinks unconditionally because I can imagine lots
			# of weirdness happening with them otherwise and they don't
			# take up much space.
			shippath='%s/shipping/%s/%s'%(todir,sub,file)
			os.link(topath,shippath)
			os.unlink(topath)
		    elif os.path.isfile(topath):
			# Ok, it's a file and previously existed..make an xdelta.
			deltapath='%s/delta/%s/%s'%(todir,sub,file)
			if os.system("xdelta delta -9 %s %s %s"%(frompath,topath,deltapath)):
			    print "Ignoring (probably gratuitous) xdelta error. Please go bug the xdelta author to\nfix this :)"
			    #raise DeltaException,"Could not create patchfile for /%s/%s"%(sub,file)
			if os.path.getsize(deltapath) >= os.path.getsize(topath):
			    print "Delta for %s/%s is larger than the file itself, moving to shipping"%(sub,file)
			    os.unlink(deltapath)
			    os.link(topath,'%s/shipping/%s/%s'%(todir,sub,file))
			    os.unlink(topath)
			else:
			    # Preserve permissions!
			    mode=os.stat(topath)[0]
			    os.chmod(deltapath,mode)
		    else:
			# Erm.  Maybe we found a FIFO or something? .. I .. I dunno..
			# Bail! :)  FIXME: do something intelligent
			raise "Panic","I don't know what to do with /%s/%s!"%(sub,file)
	walk('',fromdir,todir)

	print "Creating patchfile:"

	print "--stamp"

	# Create the new archive.  First, create the stamp:
	open('%s/debian-diff'%todir,'w').write('%s\n'%diffver)

	# Now munge the control files to indicate patchiness [also save this
	# information for later]
	# FIXME: do this in native Python -- the correct way is to add a
	#       "Patches" key to the control dictionary and implement outputting
	#       of control dictionaries..easy EXCEPT for description fields and
	#       the fact that some fields *have* to come first..

	print "--control files"

	if os.system("sed '/^Patches:/d'  < %s/control/control | sed 's/^Description:/Patches: %s\\\nDescription:/' > %s/control && cat %s/control && mv %s/control %s/control/control"%(todir,fromfile["Version"],dir,dir,dir,todir)):
	    raise ArchiveCreationException,"Couldn't add Patches: item to package control field"

	print "--data.tar.gz"

	# Make the targz files:
	if os.system("cd %s/shipping && tar czf ../data.tar.gz *"%todir):
	    raise ArchiveCreationException,"Couldn't create data.tar.gz"

	print "--control.tar.gz"

	if os.system("cd %s/control && tar czf ../control.tar.gz *"%todir):
	    raise ArchiveCreationException,"Couldn't create control.tar.gz"

	print "--delta.tar.gz"

	if os.system("cd %s/delta && tar czf ../delta.tar.gz *"%todir):
	    raise ArchiveCreationException,"Couldn't create delta.tar.gz"

	print "--moves"

	movefile=open('%s/moves'%todir,'w')
	for i in moves:
	    dests=i[-1]
	    movefile.write('%s %s %s'%i[:-1])
	    for file in dests:
		print file
		movefile.write(' %s %i'%file)
		# Note that we include the permissions in the data written..
	    movefile.write('\n')
	movefile.close()

	if fromfile.has_key('Architecture'):
	    fromarch='_%s'%fromfile['Architecture']
	else:
	    fromarch=''

	if tofile.has_key('Architecture'):
	    toarch='_%s'%tofile['Architecture']
	else:
	    toarch=''

	outfile='%s_%s%s:%s_%s%s.deb-diff'%(fromfile['Package'],fromfile['Version'],fromarch,tofile['Package'],tofile['Version'],toarch)

	print "--%s"%outfile

	if os.path.isdir(dest):
	    outfile='%s/%s'%(dest,outfile)
	else:
	    outfile=dest

	if os.system('ar rc %s %s/debian-diff %s/control.tar.gz %s/data.tar.gz %s/delta.tar.gz %s/moves'%(outfile,todir,todir,todir,todir,todir)):
	    raise ArchiveCreationException,'Couldn\'t create %s'%outfile
    finally:
	# Clean up
	shutil.rmtree(dir)

    return debdiff(outfile)
